#!/usr/bin/env python3
#
# Copyright (c) 2006 Verdens Gang AS
# Copyright (c) 2006-2015 Varnish Software AS
# All rights reserved.
#
# Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
#
# SPDX-License-Identifier: BSD-2-Clause
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.

'''
   Generate various .c and .h files for the VCL compiler and the interfaces
   for it.
'''

import sys
import os.path

stv_variables = (
    ('free_space', 'BYTES', "0", 'storage.<name>.free_space' ),
    ('used_space', 'BYTES', "0", 'storage.<name>.used_space' ),
    ('happy',      'BOOL',  "1", 'storage.<name>.happy' ),
)

def file_header(fo):
    '''
       Header for all generated C files
    '''

    fo.write('/*\n')
    fo.write(' * NB:  This file is machine generated, DO NOT EDIT!\n')
    fo.write(' *\n')
    fo.write(' * Edit and run lib/libvcc/generate.py instead.\n')
    fo.write(' */\n')

def emit_strings(fo, what, *args):
    '''
       Spit out C-code that outputs the args with VSB_cat()

       The main job here is to escape stuff properly into
       C-language string constants, but for historical reasons
       we also wrap lines before the become too long and split
       C-strings before they get too long.
    '''

    w = 8+80+2          # Width of lines, after white space prefix
    maxlen = 20<<10     # Max length of string literal

    x = 0
    l = 0

    fo.write('\tVSB_cat(sb, "/* ---===### %s ###===--- */\\n\\n");\n' % what)
    for c in "\n".join(args):
        if l == 0:
            fo.write("\tVSB_cat(sb,\n\t    \"")
            l += 12
            x += 12
        if x == 0:
            fo.write("\t    \"")
        d = c
        if c == '\n':
            d = "\\n"
        elif c == '"':
            d = "\\\""
        elif c == '\\':
            d = "\\\\"

        if c == '\n':
            fo.write(d + "\"\n")
            x = 0
            continue
        if x > w - 2:
            fo.write(d + "\"\n")
            x = 0
            continue

        fo.write(d)
        x += len(d)
        l += len(d)
        if l > maxlen:
            fo.write("\");\n")
            l = 0
            x = 0
        if x > w - 3:
            fo.write("\"\n")
            x = 0
    if x != 0:
        fo.write("\"\n")
    if l != 0:
        fo.write("\t);\n")
    fo.write('\tVSB_cat(sb, "\\n");\n')

def emit_c_file(fo, fd, bn):
    '''
       Read a C-source file and spit out code that outputs it with VSB_cat()
    '''

    fn = os.path.join(fd, bn)

    with open(fn, encoding="utf8") as fi:
        fc = fi.read()

    fo.write("\n\t/* %s */\n\n" % fn)
    emit_strings(fo, bn, fc)

def emit_lines(fo, *args):
    '''
       Split lines into file
    '''

    for i in args:
        fo.write(i + '\n')

class Tokens():

    '''
       Everything related to VCL language tokens
       -----------------------------------------
    '''

    tokens = {
        "T_INC":        "++",
        "T_DEC":        "--",
        "T_CAND":       "&&",
        "T_COR":        "||",
        "T_LEQ":        "<=",
        "T_EQ":         "==",
        "T_NEQ":        "!=",
        "T_GEQ":        ">=",
        "T_SHR":        ">>",
        "T_SHL":        "<<",
        "T_INCR":       "+=",
        "T_DECR":       "-=",
        "T_MUL":        "*=",
        "T_DIV":        "/=",
        "T_NOMATCH":    "!~",

        # Single char tokens, for convenience on one line
        None:           "{}()*+-/%><=;!&.|~,",

        # These have handwritten recognizers
        "ID":           None,
        "CNUM":         None,
        "FNUM":         None,
        "CSTR":         None,
        "CSRC":         None,
        "CBLOB":        None,

        # End of token list
        "EOI":      None,
    }

    def __init__(self):

        # Expand single char tokens
        st = self.tokens[None]
        del self.tokens[None]
        for i in st:
            self.tokens["'" + i + "'"] = i

    def emit_vcc_token_defs(self, fn):
        '''
           Emit #define values for the tokens
        '''

        with open(fn, "w", encoding="utf8") as fo:
            file_header(fo)
            for n, i in enumerate(sorted(i for i in self.tokens if i[0] != "'")):
                assert n < 128
                fo.write("#define\t%s %d\n" % (i, 128 + n))

    def emit_vcl_tnames(self, fo):
        '''
           Emit the vcl_tnames (token->string) conversion array
        '''

        fo.write("\nconst char * const vcl_tnames[256] = {\n")
        for i, j in sorted(self.tokens.items()):
            if j is None or i[0] == "'":
                fo.write("\t[%s] = \"%s\",\n" % (i, i))
            else:
                fo.write("\t[%s] = \"%s\",\n" % (i, j))
        fo.write("};\n")

    def emit_vcl_fixed_token(self, fo):
        '''
           Emit a function to recognize tokens in a string
        '''

        fo.write('#define M1()\tdo {*q = p + 1; return (p[0]); } while (0)\n')
        fo.write('#define M2(c,t)\tdo {if (p[1] == (c)) { *q = p + 2; return (t); }} while (0)\n')

        fo.write('\nunsigned\n')
        fo.write('vcl_fixed_token(const char *p, const char **q)\n')
        fo.write('{\n')
        fo.write('\tswitch (p[0]) {\n')

        last_initial = None
        for i,j in sorted(((y, x) for x,y in self.tokens.items() if y)):
            if i[0] != last_initial:
                if last_initial:
                    fo.write("\t\tM1();\n")
                fo.write("\tcase '%s':\n" % i[0])
                last_initial = i[0]
            if len(i) == 2:
                fo.write("\t\tM2('%s', %s);\n" % (i[1], j))
        fo.write("\t\tM1();\n")
        fo.write("\tdefault:\n\t\treturn (0);\n\t}\n}\n")


def main():

    srcroot = "../.."
    buildroot = "../.."
    if len(sys.argv) == 3:
        srcroot = sys.argv[1]
        buildroot = sys.argv[2]
    elif len(sys.argv) != 1:
        print("Two arguments or none")
        exit(2)

    tokens = Tokens()

    #######################################################################
    # Our methods and actions

    returns = []

    def parse_subs(ln, c):
        ctx = ['C', 'B', 'H']
        rets = []
        for i in ln[2:]:
            if (len(i) > 4 and i.startswith('  | ``') and i.endswith('``')):
                rets.append(i[6:-2].split('(')[0])
        returns.append((ln[0][4:], ctx[c], rets))

    def parse_subs_doc(fn):
        l = []
        j = -1
        start = False
        for i in open(fn, encoding="utf8"):
            if i.startswith('##########'):
                j += 1
                l.append([])
                start = True
                continue
            if start:
                l[j].append(i.rstrip())
        for i in range(0, len(l)):
            for n in range(0, len(l[i])):
                if len(l[i][n]) < 5 or not l[i][n].startswith("~~~~~"):
                    continue
                m = n + 1
                while m < len(l[i]) and (len(l[i][m]) < 5 or not l[i][m].startswith("~~~~~")):
                    m += 1
                if m != len(l[i]):
                    m -= 2
                parse_subs(l[i][n-1:m], i)

    parse_subs_doc(os.path.join(srcroot, "doc/sphinx/reference/vcl_step.rst"))

    #######################################################################
    # Variables available in sessions
    #
    # 'all' means all methods
    # 'client' means all methods tagged "C"
    # 'backend' means all methods tagged "B"
    # 'both' means all methods tagged "B" or "C"

    varprotos = {}

    def varproto(s):
        if not s in varprotos:
            fh.write(s + ";\n")
            varprotos[s] = True

    def var_is_wildcard(sym):
        return sym[-2:] == '.*'

    def var_symbol_name(sym):
        if var_is_wildcard(sym):
            return sym[:-2]
        return sym

    class VarDef():
        def __init__(self, sym, typ, rd, wr, wu, al, vlo, vhi):
            self.sym = sym
            self.nam = var_symbol_name(sym)
            self.typ = typ
            self.rd = rd
            self.wr = wr
            self.uns = wu
            self.al = al
            self.vlo = vlo
            self.vhi = vhi

            if al is None:
                self.emit_var()
            else:
                self.emit_alias()

        def emit_var(self):
            fh.write("\n")
            fo.write("\n")
            cnam = self.nam.replace(".", "_")
            ctyp = vcltypes[self.typ]

            # fo.write("\t{ \"%s\", %s,\n" % (nm, self.typ))
            fo.write("\tsym = VCC_MkSym(tl, \"%s\", SYM_MAIN," % self.nam)
            if var_is_wildcard(self.sym):
                fo.write(" SYM_NONE, %d, %d);\n" % (self.vlo, self.vhi))
                fo.write("\tAN(sym);\n")
                fo.write("\tsym->wildcard = vcc_Var_Wildcard;\n")
            else:
                fo.write(" SYM_VAR, %d, %d);\n" % (self.vlo, self.vhi))
            fo.write("\tAN(sym);\n")
            fo.write("\tsym->type = %s;\n" % self.typ)
            if self.typ == "HEADER" and not var_is_wildcard(self.sym):
                fo.write("\tsym->eval = vcc_Eval_ProtectedHeader;\n")
            else:
                fo.write("\tsym->eval = vcc_Eval_Var;\n")

            if var_is_wildcard(self.sym):
                fo.write('\tsym->rname = "HDR_')
                fo.write(self.nam.split(".")[0].upper())
                fo.write('";\n')
            elif self.rd and self.typ != "HEADER":
                fo.write('\tsym->rname = "VRT_r_%s(ctx)";\n' % cnam)
                varproto("VCL_" + self.typ + " VRT_r_%s(VRT_CTX)" % cnam)
            fo.write("\tsym->r_methods =\n")
            restrict(fo, self.rd)
            fo.write(";\n")

            if var_is_wildcard(self.sym):
                fo.write('\tsym->lname = "HDR_')
                fo.write(self.nam.split(".")[0].upper())
                fo.write('";\n')
            elif self.wr:
                fo.write('\tsym->lname = "VRT_l_%s(ctx, ";\n' % cnam)
                s = "void VRT_l_%s(VRT_CTX, " % cnam
                if self.typ == "STRING":
                    s += ctyp.c + ", VCL_STRANDS)"
                elif self.typ == "BODY":
                    s += "enum lbody_e, const char *, " + ctyp.c + ")"
                else:
                    s += "VCL_" + self.typ + ")"
                varproto(s)
            fo.write("\tsym->w_methods =\n")
            restrict(fo, self.wr)
            fo.write(";\n")

            if self.uns:
                varproto("void VRT_u_%s(VRT_CTX)" % cnam)
                fo.write('\tsym->uname = "VRT_u_%s(ctx)";\n' % cnam)
            fo.write('\tsym->u_methods =\n')
            restrict(fo, self.uns)
            fo.write(";\n")

        def emit_alias(self):
            var_aliases.append("\tsym = VCC_MkSymAlias(tl, \"%s\", \"%s\");\n" %
                    (self.nam, self.al))
            var_aliases.append("\tAN(sym);\n")

    def parse_vcl(x):
        vlo, vhi = (0, 99)
        x = x.split()
        if x[0] == "VCL" and x[1] == "<=":
            vhi = int(float(x[2]) * 10)
        elif x[0] == "VCL" and x[1] == ">=":
            vlo = int(float(x[2]) * 10)
        else:
            print("Unknown variable version spec")
            print("XXX", x, vlo, vhi)
            exit(2)
        return vlo, vhi

    def parse_var(ln):
        l1 = ln.pop(0).split("``")
        assert len(l1) in (1, 3)
        vn = l1[0].strip()
        if len(l1) == 3:
            vlo, vhi = parse_vcl(l1[1])
        else:
            vlo, vhi = 0, 99
        vr = []
        vw = []
        vu = []
        va = None
        while True:
            l = ln.pop(0)
            if l == "":
                continue
            j = l.split()
            if j[0] == "Type:":
                assert len(j) == 2
                vt = j[1]
                continue
            if j[0] == "Readable" and j[1] == "from:":
                for i in j[2:]:
                    vr.append(i.strip(",."))
                continue
            if j[0] == "Writable" and j[1] == "from:":
                for i in j[2:]:
                    vw.append(i.strip(",."))
                continue
            if j[0] == "Unsettable" and j[1] == "from:":
                for i in j[2:]:
                    vu.append(i.strip(",."))
                continue
            if j[0] == "Alias" and j[1] == "of:":
                va = j[2]
                continue
            break
        if vn[:8] != "storage.":
            VarDef(vn, vt, vr, vw, vu, va, vlo, vhi)

    def parse_var_doc(fn):
        l = []
        with open(fn, encoding="utf8") as fi:
            for i in fi:
                l.append(i.rstrip())
        for n,x in enumerate(l):
            j = x.split()
            if len(j) != 2 or j[0] != "Type:" or not x[0].isspace():
                continue
            m = n
            while m < len(l) and (l[m] == "" or l[m][0].isspace()):
                m += 1
            parse_var(l[n-2:m-1])

    #######################################################################
    # VCL to C type conversion

    vcltypes = {}

    class VclType():
        def __init__(self, name, ctype, internal=False):
            self.name = name
            self.c = ctype
            self.internal = internal
            vcltypes[name] = self


    VclType("STRINGS", "void", True)
    VclType("SUB", "void*", True)

    with open(os.path.join(srcroot, "include/vrt.h"), encoding="utf8") as fi:
        for i in fi:
            j = i.split()
            if len(j) < 3:
                continue
            if j[0] != "typedef":
                continue
            if j[-1][-1] != ";":
                continue
            if j[-1][-2] == ")":
                continue
            if j[-1][:4] != "VCL_":
                continue
            d = " ".join(j[1:-1])
            VclType(j[-1][4:-1], d)

    #######################################################################
    # Nothing is easily configurable below this line.
    #######################################################################

    #######################################################################
    def lint_start(fo):
        fo.write('/*lint -save -e525 -e539 */\n\n')

    def lint_end(fo):
        fo.write('\n/*lint -restore */\n')

    #######################################################################

    tokens.emit_vcc_token_defs(os.path.join(buildroot, "lib/libvcc/vcc_token_defs.h"))

    #######################################################################

    rets = {}
    vcls = []
    vcls_client = []
    vcls_backend = []
    for i in returns:
        vcls.append(i[0])
        for j in i[1]:
            if j == "B":
                vcls_backend.append(i[0])
            elif j == "C":
                vcls_client.append(i[0])
        for j in i[2]:
            rets[j] = True

    #######################################################################

    with open(os.path.join(buildroot, "include/tbl/vcl_returns.h"), "w", encoding="utf8") as fo:

        file_header(fo)

        lint_start(fo)

        fo.write("#ifdef VCL_RET_MAC\n")
        ll = sorted(returns)
        for i in sorted(rets.keys()):
            fo.write("VCL_RET_MAC(%s, %s" % (i.lower(), i.upper()))
            s = ",\n\t"
            for j in ll:
                if i in j[2]:
                    fo.write("%sVCL_MET_%s" % (s, j[0].upper()))
                    s = " |\n\t"
            fo.write("\n)\n\n")
        fo.write("#undef VCL_RET_MAC\n")
        fo.write("#endif\n")

        fo.write("\n#ifdef VCL_MET_MAC\n")
        for i in ll:
            fo.write("VCL_MET_MAC(%s, %s, %s," %
                     (i[0].lower(), i[0].upper(), i[1]))
            p = " (\n\t"
            for j in sorted(i[2]):
                fo.write("%s(1U << VCL_RET_%s)" % (p, j.upper()))
                p = " |\n\t"
            fo.write(")\n)\n\n")
        fo.write("#undef VCL_MET_MAC\n")
        fo.write("#endif\n")
        lint_end(fo)

    #######################################################################

    with open(os.path.join(buildroot, "include/vcl.h"), "w", encoding="utf8") as fo:

        file_header(fo)

        emit_lines(
            fo,
            '',
            '#ifdef VCL_H_INCLUDED',
            '#  error "vcl.h included multiple times"',
            '#endif',
            '#define VCL_H_INCLUDED',
            '',
            '#ifndef VRT_H_INCLUDED',
            '#  error "include vrt.h before vcl.h"',
            '#endif',
        )

        def tbl40(a, b):
            while len(a.expandtabs()) < 40:
                a += "\t"
            return a + b

        fo.write("\n/* VCL Methods */\n")
        task = {}
        n = 1
        for i in returns:
            fo.write(tbl40("#define VCL_MET_%s" % i[0].upper(), "(1U << %d)\n" % n))
            if not i[1] in task:
                task[i[1]] = []
            task[i[1]].append("VCL_MET_" + i[0].upper())
            n += 1

        fo.write("\n" + tbl40("#define VCL_MET_MAX", "%d\n" % n))
        fo.write("\n" + tbl40("#define VCL_MET_MASK", "0x%x\n" % ((1 << n) - 1)))

        fo.write("\n")
        for i in sorted(task.keys()):
            fo.write(tbl40("#define VCL_MET_TASK_%s" % i.upper(),
                           "( " + (" | \\\n\t\t\t\t\t  ").join(task[i]) + " )\n"))

        fo.write("\n")
        fo.write(tbl40("#define VCL_MET_TASK_ALL", "( VCL_MET_TASK_"))
        fo.write(" | \\\n\t\t\t\t\t  VCL_MET_TASK_".join(map(str.upper, task.keys())))
        fo.write(" )")

        fo.write("\n/* VCL Returns */\n")
        n = 1
        for i in sorted(rets.keys()):
            fo.write(tbl40("#define VCL_RET_%s" % i.upper(), "%d\n" % n))
            n += 1

        fo.write("\n" + tbl40("#define VCL_RET_MAX", "%d\n" % n))

        fo.write("\n/* VCL Types */\n")
        emit_lines(
            fo,
            'struct vrt_type {',
            '\tunsigned\t\tmagic;',
            '#define VRT_TYPE_MAGIC\t\t0xa943bc32',
            '\tconst char\t\t*lname;',
            '\tconst char\t\t*uname;',
            '\tconst char\t\t*ctype;',
            '\tsize_t\t\t\tszof;',
            '};',
            '',
        )

        for vcltype in sorted(vcltypes.keys()):
            fo.write("extern const struct vrt_type VCL_TYPE_%s[1];\n" % vcltype)

        emit_lines(
            fo,
            '',
            '/* Compiled VCL Interface */',
            'typedef int vcl_event_f(VRT_CTX, enum vcl_event_e);',
            'typedef int vcl_init_f(VRT_CTX);',
            'typedef void vcl_fini_f(VRT_CTX);',
            '',
            'struct VCL_conf {',
    	    '	unsigned		magic;',
            '#define VCL_CONF_MAGIC		0x7406c509      /* from /dev/random */',
            '',
    	    '	unsigned		syntax;',
    	    '	VCL_BACKEND		*default_director;',
    	    '	VCL_PROBE		default_probe;',
    	    '	unsigned		nref;',
    	    '	const struct vpi_ref	*ref;',
            '',
    	    '	unsigned		nsrc;',
    	    '	unsigned		nsub;',
    	    '	const char		**srcname;',
    	    '	const char		**srcbody;',
            '',
    	    '	int			nvmod;',
    	    '	const struct vpi_ii	*instance_info;',
            '',
    	    '	vcl_event_f		*event_vcl;',
            '',
        )

        for i in returns:
            fo.write("\tvcl_func_f\t\t*" + i[0] + "_func;\n")

        fo.write("\n};\n")
    fo.close()

    #######################################################################

    with open(os.path.join(buildroot, "include/tbl/vcl_context.h"), "w", encoding="utf8") as fo:
        file_header(fo)

        fo.write("/*lint -save -e525 -e539 */\n")
        for i in returns:
            fo.write("\nVCL_CTX(vcl_%s,%s)" % (i[0],i[0].upper()))
        fo.write("\nVCL_CTX(backend, TASK_B)")
        fo.write("\nVCL_CTX(client, TASK_C)")
        fo.write("\nVCL_CTX(housekeeping, TASK_H)")
        fo.write("/*lint -restore */\n")
        fo.write("\n")
        fo.write("\n#undef VCL_CTX")
        fo.write("\n")

    #######################################################################

    def restrict(fo, spec):
        d = {}
        for j in spec:
            if j[:4] == "vcl_":
                j = j[4:]
            if j == 'all':
                for i in vcls:
                    d[i] = True
            elif j == 'backend':
                for i in vcls_backend:
                    d[i] = True
            elif j == 'client':
                for i in vcls_client:
                    d[i] = True
            elif j == 'both':
                for i in vcls_client:
                    d[i] = True
                for i in vcls_backend:
                    d[i] = True
            else:
                if not j in vcls:
                    print("JJ", j)
                assert j in vcls
                d[j] = True
        p = ""
        l = list(d.keys())
        l.sort()
        w = 0
        fo.write("\t\t")
        for j in l:
            x = p + "VCL_MET_" + j.upper()
            if w + len(x) > 60:
                fo.write("\n\t\t")
                w = 0
            fo.write(x)
            w += len(x)
            p = " | "
        if not d:
            fo.write("0")

    #######################################################################

    with open(os.path.join(buildroot, "include/vrt_obj.h"), "w", encoding="utf8") as fh:
        with open(os.path.join(buildroot, "lib/libvcc/vcc_obj.c"), "w", encoding="utf8") as fo:
            file_header(fh)
            file_header(fo)

            fo.write('\n#include "config.h"\n\n#include "vcc_compile.h"\n')
            fo.write('\nvoid\nvcc_Var_Init(struct vcc *tl)\n{\n\tstruct symbol *sym;\n')

            var_aliases = []
            parse_var_doc(os.path.join(srcroot, "doc/sphinx/reference/vcl_var.rst"))
            for al in var_aliases:
                fo.write(al)
            fo.write("}\n")

            for i in stv_variables:
                fh.write("VCL_" + i[1] + " VRT_stevedore_" + i[0] + "(VCL_STEVEDORE);\n")

            fo.write("\n/* VCL type identifiers */\n")

            for vn in sorted(vcltypes.keys()):
                v = vcltypes[vn]
                if v.internal:
                    continue
                fo.write("const struct vrt_type VCL_TYPE_%s[1] = { {\n" % v.name)
                fo.write("\t.magic = VRT_TYPE_MAGIC,\n")
                fo.write('\t.lname = "%s",\n' % v.name.lower())
                fo.write('\t.uname = "%s",\n' % v.name)
                fo.write('\t.ctype = "%s",\n' % v.c)
                if v.c != "void":
                    fo.write('\t.szof = sizeof(VCL_%s),\n' % v.name)
                fo.write("}};\n")

    #######################################################################

    with open(os.path.join(buildroot, "lib/libvcc/vcc_fixed_token.c"), "w", encoding="utf8") as fo:

        file_header(fo)
        fo.write('\n#include "config.h"\n\n#include "vcc_compile.h"\n')

        tokens.emit_vcl_fixed_token(fo)
        tokens.emit_vcl_tnames(fo)

        fo.write('\nvoid\nvcl_output_lang_h(struct vsb *sb)\n{\n')

        emit_c_file(fo, srcroot, "include/vdef.h")
        emit_c_file(fo, srcroot, "include/vrt.h")
        emit_c_file(fo, buildroot, "include/vcl.h")
        emit_c_file(fo, buildroot, "include/vrt_obj.h")
        emit_c_file(fo, srcroot, "include/vcc_interface.h")
        emit_strings(
            fo,
            "vgc asserts (generate.py)",
            '#define assert(e) 								\\',
            '	do {									\\',
    	    '		if (!(e)) {							\\',
            '	        	VPI_Fail(__func__, __FILE__, __LINE__, #e);		\\',
    	    '		}								\\',
            '	} while (0)',
            '',
        )

        fo.write("\n}\n")

    #######################################################################

    with open(os.path.join(buildroot, "lib/libvcc/vcc_types.h"), "w", encoding="utf8") as ft:
        file_header(ft)

        lint_start(ft)

        for vcltype in sorted(vcltypes.keys()):
            ft.write("VCC_TYPE(" + vcltype + ", " + vcltype.lower() +")\n")
        ft.write("#undef VCC_TYPE\n")
        lint_end(ft)

    #######################################################################

    with open(os.path.join(buildroot, "include/tbl/vrt_stv_var.h"), "w", encoding="utf8") as fo:

        file_header(fo)
        lint_start(fo)

        for i in stv_variables:
            ct = vcltypes[i[1]]
            fo.write("VRTSTVVAR(" + i[0] + ",\t" + i[1] + ",\t")
            fo.write(ct.c + ",\t" + i[2] + ")")
            fo.write("\n")

        fo.write("#undef VRTSTVVAR\n")
        lint_end(fo)

if __name__ == "__main__":
    main()



#!/usr/bin/env python3
#
# Very basic style-checker for include/tbl files.

import glob

def check_file(fn):
    print("Check", fn)
    ll = []
    for l in open(fn):
        ll.append(l)

    assert ll.pop(0)[:2] == "/*"

    while ll.pop(0) != " */\n":
        continue

    assert len(ll) > 5

    assert ll.pop(0) == "\n"
    i = ll.pop(0)
    assert i == "/*lint -save -e525 -e539 */\n" or \
           i == "/*lint -save -e525 -e539 -e835 */\n"
    assert ll.pop(0) == "\n"

    assert ll.pop(-1) == "/*lint -restore */\n"
    assert ll.pop(-1) == "\n"

    for i in range(0, len(ll) -1):
        assert ll[i] != "\n" or ll[i+1] != "\n"
        assert ll[i] != ")\n" or ll[i+1] == "\n" or ll[i+1][0] == "#"

    m = {}
    while ll:
        i = ll.pop(0)
        if i == "\n":
            continue
        l = i.lstrip()
        if l[0] >= 'A' and l[0] <= 'Z':
            j = l.split('(')
            m[j[0]] = "Called"
            l = l.split('//')[0]
            l = l.split('/*')[0]
            l = l.rstrip()
            if l[-1] != ')':
                while ll.pop(0).strip() != ')':
                    continue
        elif l[0] == "#":
            j = l[1:].lstrip().split()
            # print("#", j[0])
            if j[0] == "define":
                m[j[1].split("(")[0].strip()] = "Defined"
            if j[0] == "undef":
                m[j[1]] = "Undef"
            while l[-2:] == "\\\n":
                l = ll.pop(0)
        else:
            pass
            # print(l)
    rv = 0
    for i in m:
        if m[i] != "Undef":
            print("ERROR", fn, i, m[i])
            rv += 1
    return rv

rv = 0
for fn in glob.glob("*.h"):
    rv += check_file(fn)
if rv != 0:
    print(rv, "Errors")
exit(rv)



#!/usr/bin/env python3
#
# Copyright (c) 2006 Verdens Gang AS
# Copyright (c) 2006-2015 Varnish Software AS
# All rights reserved.
#
# Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
#
# SPDX-License-Identifier: BSD-2-Clause
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
# Generate vcs_version.h vmod_abi.h

import subprocess
import os
import sys

srcroot = "../.."
buildroot = "../.."
if len(sys.argv) == 3:
    srcroot = sys.argv[1]
    buildroot = sys.argv[2]
elif len(sys.argv) != 1:
    print("Two arguments or none")
    exit(2)

#######################################################################
def file_header(fo):
    fo.write("""/*
 * NB:  This file is machine generated, DO NOT EDIT!
 *
 * Edit and run include/generate.py instead.
 */

""")

#######################################################################


v = subprocess.check_output([
    "git --git-dir=%s rev-parse HEAD 2>/dev/null || echo NOGIT" %
    (os.path.join(srcroot, ".git"))
    ], shell=True, universal_newlines=True).strip()

vcsfn = os.path.join(srcroot, "include", "vcs_version.h")

try:
    i = open(vcsfn).readline()
except IOError:
    i = ""

ident = "/* " + v + " */\n"

if i != ident:
    fo = open(vcsfn, "w")
    fo.write(ident)
    file_header(fo)
    fo.write('#define VCS_Version "%s"\n' % v)
    fo.close()

    for i in open(os.path.join(buildroot, "Makefile")):
        if i[:14] == "PACKAGE_STRING":
            break
    i = i.split("=")[1].strip()

    fo = open(os.path.join(srcroot, "include", "vmod_abi.h"), "w")
    file_header(fo)
    fo.write('#define VMOD_ABI_Version "%s %s"\n' % (i, v))
    fo.close()



#!/usr/bin/env python3
#-
# Copyright (c) 2006 Verdens Gang AS
# Copyright (c) 2006-2015 Varnish Software AS
# All rights reserved.
#
# Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
# Author: Martin Blix Grydeland <martin@varnish-software.com>
#
# SPDX-License-Identifier: BSD-2-Clause
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
# Generate various .c and .h files for the VSL query expression parser
# and the interfaces for it.

import sys
import copy

srcroot = "../.."
buildroot = "../.."
if len(sys.argv) == 3:
    srcroot = sys.argv[1]
    buildroot = sys.argv[2]

#######################################################################
# These are our tokens

tokens = {
    # Numerical comparisons
    "T_EQ":         "==",
    "T_NEQ":        "!=",
    "T_LEQ":        "<=",
    "T_GEQ":        ">=",

    # String comparisons
    "T_SEQ":        "eq",
    "T_SNEQ":       "ne",

    # Regular expression matching
    "T_NOMATCH":    "!~",

    # Boolean operators
    "T_AND":        "and",
    "T_OR":         "or",
    "T_NOT":        "not",

    # Miscellaneous
    None:           "<>~[]{}():,",

    # These have handwritten recognizers
    "VAL":          None,
    "EOI":          None,

    # Special
    "T_TRUE":       None,
    "VXID":         "vxid",
}

#######################################################################
# Emit a function to recognize tokens in a string

def emit_vxp_fixed_token(fo, tokens):
    recog = list()
    emit = dict()
    for i in tokens:
        j = tokens[i]
        if j is not None:
            recog.append(j)
            emit[j] = i

    recog.sort()
    rrecog = copy.copy(recog)
    rrecog.sort(key=lambda x: -len(x))

    fo.write("""
unsigned
vxp_fixed_token(const char *p, const char **q)
{

\tswitch (p[0]) {
""")
    last_initial = None
    for i in recog:
        if i[0] == last_initial:
            continue
        last_initial = i[0]
        fo.write("\tcase '%s':\n" % last_initial)
        for j in rrecog:
            if j[0] != last_initial:
                continue

            fo.write("\t\tif (")
            k = 1
            l = len(j)
            while k < l:
                fo.write("p[%d] == '%s'" % (k, j[k]))
                fo.write(" &&\n\t\t    ")
                k += 1
            fo.write("(isword(p[%d]) ? !isword(p[%d]) : 1)) {\n" % (l - 1, l))
            fo.write("\t\t\t*q = p + %d;\n" % l)
            fo.write("\t\t\treturn (%s);\n" % emit[j])
            fo.write("\t\t}\n")
        fo.write("\t\treturn (0);\n")

    fo.write("\tdefault:\n\t\treturn (0);\n\t}\n}\n")

#######################################################################
# Emit the vxp_tnames (token->string) conversion array

def emit_vxp_tnames(fo, tokens):
    fo.write("\nconst char * const vxp_tnames[256] = {\n")
    l = list(tokens.keys())
    l.sort()
    for i in l:
        j = tokens[i]
        if j is None:
            j = i
        if i[0] == "'":
            j = i
        fo.write("\t[%s] = \"%s\",\n" % (i, j))
    fo.write("};\n")

#######################################################################

def polish_tokens(tokens):
    # Expand single char tokens
    st = tokens[None]
    del tokens[None]

    for i in st:
        tokens["'" + i + "'"] = i

#######################################################################

def file_header(fo):
    fo.write("""/*
 * NB:  This file is machine generated, DO NOT EDIT!
 *
 * Edit and run lib/libvinylapi/generate.py instead
 */
""")

#######################################################################

polish_tokens(tokens)

fo = open(buildroot + "/lib/libvinylapi/vxp_tokens.h", "w")

file_header(fo)

j = 128
l = list(tokens.keys())
l.sort()
for i in l:
    if i[0] == "'":
        continue
    fo.write("#define\t%s %d\n" % (i, j))
    j += 1
    assert j < 256

fo.close()

#######################################################################

fo = open(buildroot + "/lib/libvinylapi/vxp_fixed_token.c", "w")

file_header(fo)
fo.write("""

#include "config.h"

#include <ctype.h>
#include <stdint.h>
#include <stdio.h>

#include "vdef.h"
#include "vqueue.h"
#include "vre.h"

#include "vxp.h"
""")

emit_vxp_fixed_token(fo, tokens)
emit_vxp_tnames(fo, tokens)

fo.close()



#!/usr/bin/env python3
#
# Copyright (c) 2006-2016 Varnish Software AS
# All rights reserved.
#
# Author: Guillaume Quintard <guillaume.quintard@gmail.com>
#
# SPDX-License-Identifier: BSD-2-Clause
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
# Process various vinyltest C files and output reStructuredText to be
# included in vtc(7).

import sys
import re


def parse_file(fn, cl, tl, sl):
    p = False
    section = ""
    resec = re.compile(r"\s*/?\* SECTION: ")

    try:
        # Python3
        f = open(fn, "r", encoding="UTF-8")
    except TypeError:
        # Python2
        f = open(fn, "r")

    for l in f:
        if "*/" in l:
            p = 0
        if resec.match(l):
            a = l.split()
            section = a[2]
            sl.append(section)
            cl[section] = []
            if len(a) > 3:
                tl[section] = re.sub(
                    r"^\s*/?\* SECTION: [^ ]+ +",
                    "", l)
            else:
                tl[section] = ""
            p = 1
        elif p:
            cl[section].append(re.sub(r"^\s*\* ?", "", l))
    f.close()

if __name__ == "__main__":
    cl = {}
    tl = {}
    sl = []
    for fn in sys.argv[1:]:
        parse_file(fn, cl, tl, sl)
    sl.sort()
    for section in sl:
        print(".. _vtc-" + section + ":\n")
        print(tl[section], end="")
        a = section
        c = section.count(".")
        if c == 0:
            r = "-"
        elif c == 1:
            r = "~"
        elif c == 2:
            r = "."
        else:
            r = "*"
        print(re.sub(r".", r, tl[section]), end="")
        print("".join(cl[section]))



#!/usr/bin/env python3

# Copyright 2025 UPLEX - Nils Goroll Systemoptimierung
# All rights reserved.
#
# Author: Nils Goroll <nils.goroll@uplex.de>
#
# SPDX-License-Identifier: BSD-2-Clause
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
# generate vmod_math vcc and c files from opengroup spec html
#
# to update the generated code:
#
# curl -o vmod_math_spec.html https://pubs.opengroup.org/onlinepubs/9799919799/basedefs/math.h.html
# python vmod_math_gen.py vmod_math_spec.html vmod_math.vcc vmod_math.c


import re
import sys


vcc_type_map = {
    "int": "INT",
    "double": "REAL",
    "long": "INT",
    "real-floating": "REAL",
    "const char *": "STRING",
}

def_arg_names = ["x", "y", "z"]
arg_names_map = {
    "atan2": ["y", "x"],
    "ldexp": ["x", "e"],
    "nan": ["tag"],
    "scalbln": ["x", "e"],
    "yn": ["n", "x"],
}

# type name
re_arg = re.compile(r"^(.+?)(\s+[xy])?$")
# argument separator
re_argsep = re.compile(r",\s*")


def vcc_arg(arg, n):
    match = re_arg.search(arg)
    if match is None:
        print(f"Error: no re_arg match on '{arg}'")
        sys.exit(1)
    # print(match)

    atype = vcc_type_map[match.group(1)]
    name = match.group(2)

    if name is None:
        name = n

    return f"{atype} {name}"


def gen_func_vcc(match, vcc_fh):
    rtype = match.group(1)
    name = match.group(2)
    args = match.group(3)

    # print(match)

    rtype = vcc_type_map[rtype]
    args = re_argsep.split(args)

    argnames = def_arg_names[: len(args)]
    if name in arg_names_map:
        argnames = arg_names_map[name]

    args = ", ".join([vcc_arg(arg, name) for arg, name in zip(args, argnames)])

    tpl = f"""
$Function {rtype} {name}({args})
"""

    vcc_fh.write(tpl)


def c_arg(arg, n):
    match = re_arg.search(arg)
    if match is None:
        print(f"Error: no re_arg match on '{arg}'")
        sys.exit(1)
    # print(match)

    atype = "VCL_" + vcc_type_map[match.group(1)]
    name = match.group(2)

    if name is None:
        name = n

    return f"{atype} {name}"


def gen_func_c(match, c_fh):
    rtype = match.group(1)
    name = match.group(2)
    args = match.group(3)

    # print(match)

    rtype = "VCL_" + vcc_type_map[rtype]
    args = re_argsep.split(args)

    argnames = def_arg_names[: len(args)]
    if name in arg_names_map:
        argnames = arg_names_map[name]

    args = ", ".join([c_arg(arg, name) for arg, name in zip(args, argnames)])
    argnames = ", ".join(argnames)

    tpl = f"""
{rtype}
vmod_{name}(VRT_CTX, {args})
{{
	(void)ctx;
	return ({name}({argnames}));
}}
"""

    c_fh.write(tpl)


def gen_func(func, vcc_fh, c_fh):
    gen_func_vcc(func, vcc_fh)
    gen_func_c(func, c_fh)


# rtype name (args...)
re_spec_funcs = re.compile(r"\b([a-z0-9 ]+?)\s+([a-z0-9]+)\((.*)\)(?=;)")
# disregard variants for types we do not need or can not support
# not support: frexp(double, int *) etc returning two values
re_func_disregard = re.compile(r"long double|float\b|long long|int \*|double \*")


def gen_funcs(spec, vcc_fh, c_fh):
    header = """

math.h functions
----------------

The semantics of functions mapping directly to `math.h(7)` are not documented
herein, see the system documentation instead (for example using ``man
<function>``).

"""
    vcc_fh.write(header)

    for match in re_spec_funcs.finditer(spec):
        f = match.group(0)
        # print("XXX " + match.group(0))
        if re_func_disregard.search(f):
            # print(f"disregard: {f}")
            continue
        gen_func(match, vcc_fh, c_fh)


def gen_enumfunc(ret, func, enums, doc, vcc_fh, c_fh):
    enumcomma = ", ".join(enums)

    tpl = f"""
$Function {ret} {func}(ENUM {{ {enumcomma} }} name)

{doc}
"""

    vcc_fh.write(tpl)

    pre = f"""
VCL_{ret}
vmod_{func}(VRT_CTX, VCL_ENUM name)
{{
    (void)ctx;
"""

    post = f"""
    WRONG("{func} enum");
}}
"""

    c_fh.write(pre)
    for enum in enums:
        body = f"""
    if (name == VENUM({enum})) return ({enum});
"""
        c_fh.write(body)
    c_fh.write(post)


re_spec_const = re.compile(r">(M_[A-Z0-9_]+)<")
# Austin Group Defect 1503 - not yet available on Linux as of Debian 12.12 2025-12
re_const_disregard = re.compile("^(M_1_SQRTPI|M_EGAMMA|M_PHI|M_SQRT1_3|M_SQRT3)$")


def gen_const(spec, vcc_fh, c_fh):
    a = [
        # DBL_* from float.h
        "DBL_MANT_DIG",
        "DBL_DIG",
        "DBL_MIN_EXP",
        "DBL_MIN_10_EXP",
        "DBL_MAX_EXP",
        "DBL_MAX_10_EXP",
        "DBL_MAX",
        "DBL_EPSILON",
        "DBL_MIN",
        # HUGE_VAL not worth extracting from html
        "HUGE_VAL",
    ]

    for match in re_spec_const.finditer(spec):
        m = match.group(1)
        if re_const_disregard.search(m):
            continue
        a.append(m)

    doc = """
Return the value of the named constant. For ``DBL_*`` see `float.h(7)`,
otherwise `math.h(7)` for details.
"""
    gen_enumfunc("REAL", "constant", a, doc, vcc_fh, c_fh)


# not worth grepping the FP_ macros from the html
def gen_class(spec, vcc_fh, c_fh):
    # not worth extracting from html
    a = ["FP_INFINITE", "FP_NAN", "FP_NORMAL", "FP_SUBNORMAL", "FP_ZERO"]

    doc = """
Return the value of the named constant for comparisons of `math.fpclassify()`_
return values.

"""

    gen_enumfunc("INT", "fpclass", a, doc, vcc_fh, c_fh)


vcc_top = """
#-
# This document is licensed under the same conditions as Vinyl Cache itself.
# See LICENSE for details.
#
# SPDX-License-Identifier: BSD-2-Clause
#
# Author: Nils Goroll <nils.goroll@uplex.de>

$ABI vrt
$Module math 3 "VMOD wrapping math.h"

.. THIS CODE IS AUTO-GENERATED. DO NOT EDIT HERE. see vmod_math_gen.py

DESCRIPTION
===========

This VMOD wraps the functions in `math.h(7)`, provides additional utilities and
includes functions to access macros and constants from `math.h(7)` and
`float.h(7)`.

Utility functions
-----------------

.. source is in vmod_math_util.c

$Function BOOL approx(REAL a, REAL b, REAL maxDiff = 0.0, REAL maxRelDiff = 0.0)

.. _`Comparing Floating Point Numbers, 2012 Edition`: https://randomascii.wordpress.com/2012/02/25/comparing-floating-point-numbers-2012-edition/

Return true if the two numbers *a* and *b* are approximately equal per the
``AlmostEqualRelativeAndAbs`` function from `Comparing Floating Point Numbers,
2012 Edition`_ given the additional parameters *maxDiff* and *maxRelDiff*. Read
the blog post for details on why the seemingly simple question of two floating
point numbers being equal has no simple answers.

For their default values of 0, the two additional parameters are initialized to:

- *maxDiff*: ``4 * DBL_EPSILON``
- *maxRelDiff*: ``DBL_EPSILON``

$Function STRING strfromd(STRING format, REAL fp)

Convert the value *fp* into a string using *format*, see `strfromd(3)`.

An invalid format string results in a VCL error and the ``NULL`` invalid string
returned.

.. Internally, vsnprintf() is called via WS_Printf(). The difference to
   VRT_REAL_string() is the free format and missing call to VRT_REAL_is_valid()

Access to macros and constants
------------------------------

"""

c_top = """
/* this code is auto-generated. Do not edit */
#include "config.h"

#include <float.h>
#include <math.h>

#include "vdef.h"
#include "vas.h"
#include "vrt.h"
#include "vcc_math_if.h"

// macros in math.h code are generic-ish across float/double/long double
//lint --e{506} Constant value Boolean
//lint --e{736} Loss of precision
"""


def main():
    # Check command line arguments
    if len(sys.argv) != 4:
        print("Usage: python3 script.py <spec_file> <vcc_file> <c_file>")
        sys.exit(1)

    spec_file = sys.argv[1]
    vcc_file = sys.argv[2]
    c_file = sys.argv[3]

    spec = ""

    try:
        with open(spec_file, "r") as f:
            spec = f.read()
    except FileNotFoundError:
        print(f"Error: Input file '{spec_file}' not found")
        sys.exit(1)

    with open(vcc_file, "w") as vcc_fh:
        vcc_fh.write(vcc_top)
        with open(c_file, "w") as c_fh:
            c_fh.write(c_top)
            gen_const(spec, vcc_fh, c_fh)
            gen_class(spec, vcc_fh, c_fh)
            gen_funcs(spec, vcc_fh, c_fh)
    print(f"Output written to '{vcc_file}' and '{c_file}'")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
#
# Copyright (c) 2017 Varnish Software AS
# All rights reserved.
#
# Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
#
# SPDX-License-Identifier: BSD-2-Clause
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.

"""
This program produces a compact summ of gcov run on all .o files
found in a subdirectory tree.

Options:

    -g gcov-program
        default: "llvm-cov gcov"

    -o output-filename
        default: stdout

    -x exclude-subdir
        default ".git" and ".deps"

 Arguments:

    directories to process.
    default: .

"""

import os
import sys
import getopt
import subprocess

source_files = {}

class SourceFile():
    ''' GCOV counts for a single source file '''

    def __init__(self, path):
        self.path = path
        self.counts = {}
        self.length = 0
        source_files[path] = self

    def __lt__(self, other):
        return self.path < other.path

    def ingest(self, fd):
        ''' Ingest a gcov file '''
        for line in fd:
            flds = line.split(':', maxsplit=2)
            cnt = flds[0].strip()
            if cnt == "-":
                continue
            if cnt == "#####":
                cnt = 0
            else:
                cnt = int(cnt)
            lno = int(flds[1])
            if lno not in self.counts:
                self.counts[lno] = cnt
            else:
                self.counts[lno] += cnt
        self.length = lno

    def summary(self):
        """
        Produce compact output file

        Format:
            linefm [lineto] count

            "+" in linefm means "previous line + 1"
            "." in count means "same as previous count"
        """
        yield "/" + self.path + " " + str(self.length)
        lnos = list(sorted(self.counts.items()))
        pln = -1
        pcn = -1
        while lnos:
            ln, cn = lnos.pop(0)
            lnl = ln
            while lnos:
                lnn, cnn = lnos[0]
                if lnl + 1 != lnn or cnn != cn:
                    break
                lnos.pop(0)
                lnl = lnn
            if ln == pln + 1:
                s = "+ "
            else:
                s = "%d " % ln

            if ln != lnl:
                s += "%d " % lnl
                pln = lnl
            else:
                pln = ln

            if cn == pcn:
                s += "."
            else:
                s += "%d" % cn
                pcn = cn
            yield s

exclude = [".git", ".deps",]

def process_gcov(fn, root):
    """ Sum .gcov file into counts, then delete it """

    with open(fn, encoding="UTF-8") as gfile:
        sline = gfile.readline().split(':')
        assert sline[2] == "Source"
        sname = os.path.normpath(os.path.join(root, sline[3].strip()))
        sfile = source_files.get(sname)
        if not sfile:
            sfile = SourceFile(sname)
        sfile.ingest(gfile)
    os.remove(fn)

def run_gcov(prog, subdir):
    """ Walk tree, run gcov, process output """
    for root, dirs, files in os.walk(subdir):
        for i in exclude:
            if i in dirs:
                dirs.remove(i)
        if " ".join(files).find(".gcda") == -1:
            continue
        for fn in files:
            if fn[-2:] != ".o":
                continue

            # if we find the .o file in a .../.libs the sources
            # must be found relative to the parent directory

            if "vinyld" in root:
                subdir = root.split("/")[-1]
                cmd = ["cd " + root + "/.. && " + "exec " + prog + " " + subdir + "/" + fn]
                rpath = "/../"
            elif "vudp" in fn:
                cmd = ["cd " + root + "/../.. && " + "exec " + prog + " vtest2/lib/" + fn]
                rpath = "/../../"
                #print("VT2.LIB")
            elif "vtest2" in root:
                cmd = ["cd " + root + "/../.. && " + "exec " + prog + " vtest2/src/" + fn]
                rpath = "/../../"
                #print("VT2")
            elif root[-6:] == "/.libs":
                cmd = ["cd " + root + "/.. && " + "exec " + prog + " .libs/" + fn]
                rpath = "/../"
            else:
                #print("R", root, "FN", fn)
                cmd = ["cd " + root + " && " + "exec " + prog + " " + fn]
                rpath = "/"

            sp = subprocess.run(
                    cmd,
                    shell=True,
                    universal_newlines=True,
                    capture_output=True,
            )

            if sp.returncode != 0 or sp.stderr:
                print("CMD", cmd, "EXIT", sp.returncode)
                print("SERR", len(sp.stderr), sp.stderr)
                print("SOUT", sp.stdout)
                sys.exit(2)
            x = sp.stdout

            for ln in x.split("\n"):
                if "such file" in ln:
                    print("LN", ln)
                    assert "such file" not in ln
                ln = ln.split()
                if not ln:
                    continue
                if "reating" in ln[0]:
                    gn = root + rpath + ln[1].strip("'")
                    gn = os.path.normpath(gn)
                    assert gn[-5:] == ".gcov"
                    #print("GS", gn)
                    process_gcov(gn, root.replace("/.libs", ""))

def main():
    ''' ... '''
    optlist, args = getopt.getopt(sys.argv[1:], "g:o:x:")

    fo = sys.stdout
    gcovprog = os.environ.get('GCOVPROG')
    if gcovprog is None:
        gcovprog = "llvm-cov gcov"

    for f, v in optlist:
        if f == '-o' and v == '-':
            fo = sys.stdout
        elif f == '-o':
            fo = open(v, "w", encoding="utf8")
        elif f == '-g':
            gcovprog = v
        elif f == '-x':
            exclude.append(v)
        else:
            assert False
    if not args:
        args = ["."]
    for dn in args:
        run_gcov(gcovprog, dn)

    for sf in sorted(source_files.values()):
        for i in sf.summary():
            fo.write(i + "\n")

if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import os

def check(fn):
    l = []
    for i in open(fn):
        i = i.strip()
        if not i:
            continue
        if i[0] != "#":
            continue
        if i.find("include") == -1:
            continue
        if i.find('"') == -1:
            l.append(i.split('<')[1].split('>')[0])
        else:
            l.append(i.split('"')[1])
    if "vrt.h" in l:
        vrt = l.index("vrt.h")
        if "vdef.h" not in l:
            print(fn, "vdef.h not included with vrt.h")
        vdef = l.index("vdef.h")
        if vdef > vrt:
            print(fn, "vdef.h included after vrt.h")
        for i in ("stddef.h", "stdint.h", "cache/cache.h", "cache.h"):
            if i in l:
                print(fn, i + " included with vrt.h")

    for i in ("cache/cache.h", "cache.h"):
        if i in l:
            for j in ("stddef.h", "stdint.h", "vrt.h",
                      "math.h", "pthread.h", "stdarg.h", "sys/types.h",
                      "vdef.h", "miniobj.h", "vas.h", "vqueue.h", "vtree.h"):
                if j in l:
                    print(fn, j + " included with cache.h")

for (dir, dns, fns) in os.walk("."):
    for f in fns:
        if f[-2:] == ".c":
            check(dir + "/" + f)



#!/usr/bin/env python3
#
# Copyright (c) 2021 Varnish Software AS
# All rights reserved.
#
# Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
#
# SPDX-License-Identifier: BSD-2-Clause
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#

''' Figure out what developer-warnings CC can grok '''

import os
import sys
import subprocess
import tempfile

DESIRABLE_OPTIONS = [
    "-fstack-protector-strong",
    "-Werror",
    "-Wall",
]

DESIRABLE_WFLAGS = [
    "-Wcast-align",
    "-Wcast-qual",
    "-Wchar-subscripts",
    "-Wempty-body",
    "-Wextra",
    "-Wformat -Wformat-y2k",
    "-Wformat -Wformat-zero-length",
    "-Wmissing-field-initializers",
    "-Wmissing-prototypes",
    "-Wnested-externs",
    "-Wpointer-arith",
    "-Wpointer-sign",
    "-Wreturn-type",
    "-Wshadow",
    "-Wstrict-aliasing",
    "-Wstrict-prototypes",
    "-Wstring-plus-int",
    "-Wswitch",
    "-Wunused-parameter",
    "-Wunused-parameters",
    "-Wunused-result",
    "-Wunused-but-set-variable",
    "-Wwrite-strings",
]

UNDESIRABLE_WFLAGS = [
    "-Wno-system-headers", # Outside of our control
    "-Wno-thread-safety", # Does not understand our mutexes are wrapped
    "-Wno-sign-compare", # Fixable
    "-Wno-gnu-folding-constant", # does not help code clarity
]


def cc(compiler, opt, obj, src):
    a = compiler.split()
    a += ["-c"]
    if opt is not None:
        a += opt.split()
    a += ["-o", obj, src]

    try:
        j = subprocess.check_output(a, stderr=subprocess.STDOUT)
    except subprocess.CalledProcessError as err:
        if err.output:
            j = err.output
        else:
            j = ("Returncode %d" % err.returncode).encode('utf8')
    return (j)


def main():
    compiler = os.environ.get("CC", "cc")

    src_file = tempfile.NamedTemporaryFile(suffix='.c')
    src_file.write(b'int main(int argc, char **argv) {(void)argc;(void)argv;return(0);}\n')
    src_file.flush()
    obj_file = tempfile.NamedTemporaryFile(suffix='.o')

    j = cc(compiler, None, obj_file.name, src_file.name)
    if j:
        sys.stderr.write(compiler + " failed without flags\n\t" +
                         j.decode('utf8') + '\n')
        sys.exit(1)

    use_flags = []
    for i in DESIRABLE_OPTIONS + DESIRABLE_WFLAGS + UNDESIRABLE_WFLAGS:
        sys.stderr.write("checking whether C compiler accepts " + i + "... ")
        j = cc(compiler, i, obj_file.name, src_file.name)
        if not j:
            use_flags.append(i)
            sys.stderr.write("yes\n")
        else:
            sys.stderr.write(" no\n")
            if b'error: unrecognized command line option' in j:
                # LLVM
                pass
            elif b'warning: unknown warning option' in j:
                # GCC
                pass
    print(" ".join(use_flags))

if __name__ == "__main__":
    main()
